import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import Color, LocationType, LocationReference
from ablelabs.neon.utils.format_conversion import floor_decimal


class Speed:
    def __init__(self, unit: str, unit_s: float = None, rate: float = None) -> None:
        self.unit = unit
        self.unit_s = unit_s
        self.rate = rate

    def __str__(self) -> str:
        result = []
        if self.unit_s:
            result.append(f"{self.unit_s}{self.unit}/s")
        if self.rate:
            result.append(f"{floor_decimal(self.rate * 100, digit=1)}%")
        return f"({' '.join(result)})"

    @staticmethod
    def from_mm(mm: float):
        return Speed(unit="mm", unit_s=mm)

    @staticmethod
    def from_rate(rate: float = 1.0):
        return Speed(unit="mm", rate=rate)


class FlowRate(Speed):
    @staticmethod
    def from_ul(ul: float):
        return Speed(unit="ul", unit_s=ul)

    @staticmethod
    def from_rate(rate: float = 1.0):
        return Speed(unit="ul", rate=rate)


class LedBarParam:
    def __init__(
        self,
        color: Color = Color.NONE,
        on_brightness_percent: int = None,
        off_brightness_percent: int = None,
        bar_percent: int = None,
        blink_time_ms: int = None,
    ) -> None:
        self.color = color
        self.on_brightness_percent = on_brightness_percent
        self.off_brightness_percent = off_brightness_percent
        self.bar_percent = bar_percent
        self.blink_time_ms = blink_time_ms


class Location:
    def __init__(
        self,
        location_type: LocationType,
        location_number: int,
        well: str,
        reference: LocationReference,
        offset: tuple[float] = (0, 0, 0),
    ) -> None:
        self.location_type = location_type
        self.location_number = location_number
        self.well = well
        self.reference = reference
        self.offset = offset

    def to(
        self,
        location_number: int = None,
        well: str = None,
        reference: LocationReference = None,
        offset: tuple[float] = None,
    ):
        return Location(
            location_type=self.location_type,
            location_number=(
                location_number if location_number else self.location_number
            ),
            well=well if well else self.well,
            reference=reference if reference else self.reference,
            offset=offset if offset else self.offset,
        )

    def __str__(self) -> str:
        # inspect만으로는 원하는 형태로 만들기가 어려울 듯.
        # result = " ".join(
        #     [
        #         f"{name}={value}"
        #         for name, value in inspect.getmembers(self)
        #         if "__" not in name and not inspect.isfunction(value)
        #     ]
        # )
        # return result
        result = f"{self.location_type}"
        if self.location_number:
            result += f".{self.location_number}"
        if self.well:
            result += f" well={self.well}"
        if self.reference:
            result += f" reference={self.reference}"
        if self.offset and self.offset != (0, 0, 0):
            result += f" offset={self.offset}"
        return result
    
    def __repr__(self) -> str:
        return self.__str__()


if __name__ == "__main__":
    print(
        Location(
            location_type=LocationType.DECK,
            location_number=12,
            well="a6",
            reference=LocationReference.BOTTOM,
            offset=(1, 2, 3),
        )
    )



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.suitable.sara.robot_router import RobotRouter
from ablelabs.neon.common.suitable.enums import Axis


class AxisAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.axis_get_position)
    async def get_position(self, axis: Axis, floor_digit: int = 1):
        pass

    @run_server_func(RobotRouter.axis_set_speed)
    async def set_speed(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_set_accel)
    async def set_accel(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_set_decel)
    async def set_decel(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_enable)
    async def enable(self, axis: Axis):
        pass

    @run_server_func(RobotRouter.axis_disable)
    async def disable(self, axis: Axis):
        pass

    @run_server_func(RobotRouter.axis_stop)
    async def stop(self, axis: Axis):
        pass

    @run_server_func(RobotRouter.axis_home)
    async def home(self, axis: Axis):
        pass

    @run_server_func(RobotRouter.axis_jog)
    async def jog(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_step)
    async def step(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_move)
    async def move(self, axis: Axis, value: float):
        pass

    @run_server_func(RobotRouter.axis_wait_home_done)
    async def wait_home_done(self, axis: Axis):
        pass

    @run_server_func(RobotRouter.axis_wait_move_done)
    async def wait_move_done(self, axis: Axis):
        pass



from typing import Any, Callable
import asyncio
from functools import update_wrapper

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.decorators import (
    log_func_args_async,
    get_args_dict,
    get_func_args_str,
)


class Work:
    def __init__(self, func: Callable, args: dict) -> None:
        self.func = func
        self.args = args

    def __str__(self) -> str:
        return get_func_args_str(self.func, self.args)

    async def execute(self):
        func: Callable = self.func
        args = self.args
        try:
            if isinstance(args, dict):
                return await func(**args)
            elif isinstance(args, tuple):
                return await func(*args)
            else:
                return await func(args)
        # to_ 함수인 경우, 리턴이 list라서 await가 안 걸려서 TypeError 발생.
        except TypeError as e:
            pass


class WorkNode:
    def __init__(
        self,
        work: Work,
        children: list = None,
        log_before_func: Callable = None,
        log_after_func: Callable = None,
    ) -> None:
        self.work = work
        self.children: list[WorkNode] = children or []
        self.log_before_func = log_before_func
        self.log_after_func = log_after_func

    def __str__(self) -> str:
        return f"{self.work}"

    def get_all_work_node_str(self):
        str_list = self._get_all_work_node_strs()
        return "\n".join(str_list)

    def _get_all_work_node_strs(self, depth=0):
        result = ["\t" * depth + f"{self.work}"]
        for child in self.children:
            result.extend(child._get_all_work_node_strs(depth + 1))
        return result


def to_work_node(
    origin_func: Callable = None,
    log_before_func: Callable = None,
    log_after_func: Callable = None,
):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            _func = origin_func if origin_func else func
            args_dict = get_args_dict(_func, *args, **kwargs)
            children = func(**args_dict)
            result = WorkNode(
                work=Work(_func, args_dict),
                children=children,
                log_before_func=log_before_func,
                log_after_func=log_after_func,
            )
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


def serialize(work_node: WorkNode):
    work_nodes: list[WorkNode] = (
        work_node if isinstance(work_node, list) else [work_node]
    )
    result: list[Work] = []
    for _work_node in work_nodes:
        result.append(_work_node.work)
        if childs := serialize(_work_node.children):
            if log_func := _work_node.log_before_func:
                result.append(Work(func=log_func, args=f"{_work_node.work}"))
            result.extend(childs)
            if log_func := _work_node.log_after_func:
                result.append(Work(func=log_func, args=f"{_work_node.work} -> done"))
    return result


async def execute_work_node(work_node: WorkNode | list[WorkNode]):
    works = serialize(work_node)
    for work in works:
        await work.execute()


def async_work_node_executor(work_node_method: Callable[[Any], list[WorkNode]]):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            work_node = work_node_method(*args, **kwargs)
            return await execute_work_node(work_node)

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


@log_func_args_async(log_before_func=print, log_after_func=print)
async def test_simple_func(delay: float):
    await asyncio.sleep(delay)


@to_work_node(test_simple_func)
def test_to_simple_func(delay: float) -> Work:
    pass


@to_work_node(log_before_func=print, log_after_func=print)
def test_to_complex_func(a: float, b: float, c: float):
    res: list[WorkNode] = []
    res.append(test_to_simple_func(delay=a))
    res.append(test_to_simple_func(b))
    res.append(test_to_simple_func(delay=c))
    return res


@async_work_node_executor(test_to_complex_func)
def test_complex_func(a: float, b: float, c: float):
    pass


@to_work_node(log_before_func=print, log_after_func=print)
def test_to_more_complex_func(a: float, b: float, c: float, d: float, e: float):
    res: list[WorkNode] = []
    res.append(test_to_simple_func(delay=a))
    res.append(test_to_complex_func(b, b=c, c=d))
    res.append(test_to_simple_func(e))
    return res


@async_work_node_executor(test_to_more_complex_func)
def test_more_complex_func(a: float, b: float, c: float, d: float, e: float):
    pass


async def main():
    # await test_simple_func(delay=1)
    # print(simple_work_node := test_to_simple_func(delay=1))
    # await execute_work_node(simple_work_node)

    # print(complex_work_node := test_to_complex_func(1, b=2, c=3))
    # print(await execute_work_node(complex_work_node))
    # print(
    #     more_complex_work_node := test_to_more_complex_func(
    #         a=0.1, b=0.2, c=0.3, d=0.4, e=0.5
    #     )
    # )
    # print(await execute_work_node(more_complex_work_node))

    # await test_complex_func(1, b=2, c=3)
    # await test_more_complex_func(0.1, b=0.2, c=0.3, d=0.4, e=0.5)

    # complex_work_node = test_to_complex_func(a=1, b=2, c=3)
    # print(complex_work_node)
    # await execute_work_node(complex_work_node)
    more_complex_work_node: WorkNode = test_to_more_complex_func(
        0.1, b=0.2, c=0.3, d=0.4, e=0.5
    )
    print(more_complex_work_node.get_all_work_node_str())
    await execute_work_node(more_complex_work_node)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from loguru import logger

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
# from neon.utils.decorators import log_func_args_async
from ablelabs.neon.controllers.suitable.sara.robot_router import RobotRouter
from ablelabs.neon.controllers.suitable.sara.api.set_api import SetAPI
from ablelabs.neon.controllers.suitable.sara.api.state_api import StateAPI
from ablelabs.neon.controllers.suitable.sara.api.motion_api import MotionAPI
from ablelabs.neon.controllers.suitable.sara.api.axis_api import AxisAPI
from ablelabs.neon.common.suitable.constants import PIPETTE_NUMBERS
from ablelabs.neon.common.suitable.enums import LocationType, Axis
from ablelabs.neon.common.suitable.struct import Speed, FlowRate, location


class RobotAPI(MessengerClient):
    def __init__(self) -> None:
        tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)
        super().__init__(tcp_client)
        self._set_api = SetAPI(tcp_client=tcp_client)
        self._state_api = StateAPI(tcp_client=tcp_client)
        self._motion_api = MotionAPI(tcp_client=tcp_client)
        self._axis_api = AxisAPI(tcp_client=tcp_client)

    @property
    def set(self):
        return self._set_api

    @property
    def state(self):
        return self._state_api

    @property
    def motion(self):
        return self._motion_api

    @property
    def axis(self):
        return self._axis_api

    # @log_func_args_async(log_before_func=logger.info, log_after_func=logger.debug)
    async def connect(self, ip, port):
        while True:
            try:
                await self._tcp_client.connect(ip=ip, port=port)
            except ConnectionRefusedError as e:
                await asyncio.sleep(1)
                continue
            else:
                break

    @run_server_func(RobotRouter.robot_wait_boot)
    async def wait_boot(self):
        pass

    @run_server_func(RobotRouter.robot_stop)
    async def stop(self):
        pass

    @run_server_func(RobotRouter.robot_clear_error)
    async def clear_error(self):
        pass

    @run_server_func(RobotRouter.robot_pause)
    async def pause(self):
        pass

    @run_server_func(RobotRouter.robot_resume)
    async def resume(self):
        pass

    @run_server_func(RobotRouter.robot_is_connected)
    async def is_connected(self):
        pass

    @run_server_func(RobotRouter.robot_get_environment)
    async def get_environment(self):
        pass


async def main():
    import subprocess

    subprocess.Popen(
        [r".venv\Scripts\python.exe", r"robot\src\controllers\suitable\sara\router.py"]
    )

    logger.remove()
    # logger.add(sys.stdout, level="TRACE")
    # logger.add(sys.stdout, level="DEBUG", backtrace=False)
    logger.add(sys.stdout, level="INFO", backtrace=False)
    # logger.add("logs/trace.log", level="TRACE")
    # logger.add("logs/debug.log", level="DEBUG")
    logger.add("logs/info.log", level="INFO")

    ip = "localhost"
    port = 1234

    robot_api = RobotAPI()
    await robot_api.connect(ip=ip, port=port)

    robot_api.set.pipettes({n: "1ch1000ul" for n in PIPETTE_NUMBERS})
    robot_api.set.tips({n: "tip_1000" for n in PIPETTE_NUMBERS})
    robot_api.set.labwares(
        {
            1: "bioneer_384_qc_plate",
            2: "spl_dw_reservoir",
            3: "ablelabs_tiprack_1000ul",
            4: "spl_96_deep_well_plate",
            10: "spl_96_deep_well_plate",
            16: "spl_96_deep_well_plate",
            22: "spl_96_deep_well_plate",
            5: "ablelabs_tiprack_1000ul",
            11: "ablelabs_tiprack_1000ul",
            17: "ablelabs_tiprack_1000ul",
            23: "ablelabs_tiprack_1000ul",
            6: "ablelabs_tiprack_200ul",
            12: "ablelabs_tiprack_200ul",
            18: "ablelabs_tiprack_200ul",
            24: "ablelabs_tiprack_200ul",
        }
    )

    await robot_api.stop()
    await robot_api.clear_error()  # stop 하고
    await robot_api.pause()
    await robot_api.resume()
    logger.info(f"is_connected = {await robot_api.is_connected()}")
    logger.info(f"get_environment = {await robot_api.get_environment()}")

    async def get_current_motion():
        while True:
            logger.info(
                f"get_current_motion = {await robot_api.state.get_current_motion()}"
            )

    getter_tasks = [
        asyncio.create_task(getter)
        for getter in [
            get_current_motion(),
        ]
    ]
    for getter_task in getter_tasks:
        getter_task.cancel()
    # if any([isinstance(task.result(), Exception) for task in tasks]):
    #     return

    await robot_api.motion.initialize()
    await robot_api.motion.move_to(
        pipette_number=[1, 2, 3],
        location=location(
            location_type=LocationType.DECK,
            location_number=4,
            well=["a1", "b1", "c1"],
        ),
    )
    await robot_api.motion.pick_up_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=3,
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.DECK,
            location_number=3,
            well=["a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"],
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        location=location(
            location_type=LocationType.WASTE,
        ),
    )
    # await robot_api.motion.mix(
    #     volume=100,
    #     iteration=2,
    #     location=location(location_type=LocationType.DECK, location_number=, offset=(0, 0, 5)),
    #     pipette_cone=1,
    #     flow_rate=FlowRate.from_ul(100),
    #     blow_out_flow_rate=FlowRate.from_ul(10),
    #     delay=0.1,
    # )
    # await robot_api.motion.blow_out(
    #     flow_rate=FlowRate.from_ul(50),
    # )
    await robot_api.motion.rise_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    # await robot_api.motion.aspirate(
    #     volume=200,
    #     location=location(location_type=LocationType.DECK, location_number=, offset=(0, 0, 5)),
    #     pipette_cone=1,
    #     flow_rate=FlowRate.from_ul(50),
    #     blow_out_flow_rate=FlowRate.from_ul(70),
    # )
    await robot_api.motion.delay(sec=1)
    await robot_api.motion.rise_tip(
        pipette_number=[1, 2, 3, 4, 5, 6, 7, 8],
        height_offset=10,
        z_speed=Speed.from_mm(5),
    )
    # await robot_api.motion.dispense(
    #     volume=200,
    #     location=location(location_type=LocationType.DECK, location_number=, offset=(0, 0, 10)),
    #     pipette_cone=1,
    #     flow_rate=FlowRate.from_ul(100),
    # )
    # await robot_api.motion.blow_out(
    #     flow_rate=FlowRate.from_ul(50),
    # )
    # await robot_api.motion.move_to_ready()

    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    loop.create_task(main())
    loop.run_forever()



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerServer
from ablelabs.neon.common.suitable.enums import Axis
from ablelabs.neon.common.suitable.struct import Location, Speed, FlowRate


class RobotRouter(MessengerServer):
    async def robot_wait_boot(self):
        pass

    async def robot_stop(self):
        pass

    async def robot_clear_error(self):
        pass

    async def robot_pause(self):
        pass

    async def robot_resume(self):
        pass

    async def robot_is_connected(self):
        pass

    async def robot_get_environment(self):
        pass

    # set api
    async def set_pipettes(self, value: dict[int, str | None]):
        pass

    async def set_tips(self, value: dict[int, str | None]):
        pass

    async def set_labwares(self, value: dict[int, str | None]):
        pass

    # state api
    async def state_get_current_motion(self):
        pass

    async def state_get_estimated_time(self):
        pass

    # motion api
    async def motion_initialize(self):
        pass

    async def motion_delay(self, sec: float):
        pass

    async def motion_move_to_ready(self):
        pass

    async def motion_move_to(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_pick_up_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_drop_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    async def motion_rise_tip(
        self,
        pipette_number: list[int],
        height_offset: float | list[float],
        z_speed: Speed | list[Speed],
        optimize: bool = False,
    ):
        pass

    async def motion_aspirate(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        blow_out_flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_dispense(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    async def motion_mix(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate = None,
        delay: float = 0.0,
        optimize: bool = False,
    ):
        pass

    async def motion_blow_out(
        self,
        pipette_number: list[int],
        flow_rate: FlowRate = None,
    ):
        pass

    # axis api
    async def axis_get_position(self, axis: Axis, floor_digit: int = 1):
        pass

    async def axis_set_speed(self, axis: Axis, value: float):
        pass

    async def axis_set_accel(self, axis: Axis, value: float):
        pass

    async def axis_set_decel(self, axis: Axis, value: float):
        pass

    async def axis_enable(self, axis: Axis):
        pass

    async def axis_disable(self, axis: Axis):
        pass

    async def axis_stop(self, axis: Axis):
        pass

    async def axis_home(self, axis: Axis):
        pass

    async def axis_jog(self, axis: Axis, value: float):
        pass

    async def axis_step(self, axis: Axis, value: float):
        pass

    async def axis_move(self, axis: Axis, value: float):
        pass

    async def axis_wait_home_done(self, axis: Axis):
        pass

    async def axis_wait_move_done(self, axis: Axis):
        pass



# import asyncio
# from loguru import logger

# import sys, os

# sys.path.append(os.path.abspath(os.curdir))
# from ablelabs.neon.utils.network.tcp_server import TcpServer
# from ablelabs.neon.utils.decorators import log_func_args_async
# from ablelabs.neon.controllers.suitable.sara.api.robot_router import (
#     RobotRouter as BaseRobotRouter,
# )

# # from neon.common.suitable.robot_context import *
# from ablelabs.neon.common.path import Path
# from ablelabs.neon.common.suitable.enums import LocationType, Axis
# from ablelabs.neon.common.suitable.struct import Location, Speed, FlowRate

# # from neon.common.suitable.robot_controller import *


# class RobotRouter(BaseRobotRouter):
#     def __init__(
#         self,
#         tcp_server: TcpServer,
#         robot_context: RobotContext,
#         # robot_controller: RobotController,
#     ) -> None:
#         super().__init__(tcp_server=tcp_server)
#         self._robot_context = robot_context
#         # self._robot_controller = robot_controller
#         # self._robot_command = robot_controller._robot_command
#         # self._pipette_command = robot_controller._pipette_command
#         # self._axis_command = robot_controller._axis_command
#         # self._balance_command = robot_controller._balance_command

#     # set api

#     # robot api
#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def robot_wait_boot(self):
#         await self._robot_controller.wait_boot()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.debug)
#     async def robot_stop(self):
#         await self._robot_controller.stop()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.debug)
#     async def robot_clear_error(self):
#         await self._robot_controller.clear_error()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.debug)
#     async def robot_pause(self):
#         await self._robot_controller.pause()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.debug)
#     async def robot_resume(self):
#         await self._robot_controller.resume()

#     @log_func_args_async(log_after_func=logger.debug)
#     async def robot_is_connected(self):
#         robot_driver = self._robot_controller._robot_driver
#         is_axis_connected = all(
#             motor.is_connected() for motor in robot_driver.motors.values()
#         )
#         is_balance_connected = (
#             robot_driver.balance is not None and robot_driver.balance.is_connected()
#         )
#         is_dio_connected = (
#             robot_driver.dio is not None and robot_driver.dio.is_connected()
#         )
#         return (
#             is_axis_connected,
#             is_balance_connected,
#             is_dio_connected,
#         )

#     @log_func_args_async(log_after_func=logger.debug)
#     async def robot_get_environment(self):
#         (temperature, pressure, humidity) = self._robot_controller.get_environment()
#         return (temperature, pressure, humidity)

#     # state api
#     @log_func_args_async(log_before_func=logger.debug, log_after_func=logger.info)
#     async def state_get_current_motion(self):
#         return await self._robot_controller.get_current_motion()

#     @log_func_args_async(log_before_func=logger.debug, log_after_func=logger.info)
#     async def state_get_estimated_time(self):
#         raise NotImplementedError()  # sypark

#     # motion api
#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_initialize(self):
#         await self._robot_controller.initialize()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_delay(self, sec: float):
#         await self._robot_command.delay(sec=sec)

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_move_to_ready(self):
#         await self._robot_controller.move_to_ready()

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_move_to(
#         self,
#         location: Location,
#         pipette_cone: int = 1,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.move_to(
#             location=location,
#             pipette_cone=pipette_cone,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_pick_up_tip(
#         self,
#         location: Location,
#         pipette_cone: int,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.pick_up_tip(
#             location=location,
#             pipette_cone=pipette_cone,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_drop_tip(
#         self,
#         location: Location,
#         pipette_cone: int,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.drop_tip(
#             location=location,
#             pipette_cone=pipette_cone,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_rise_tip(
#         self,
#         height_offset: float,
#         z_speed: Speed,
#     ):
#         await self._pipette_command.rise_tip(
#             height_offset=height_offset,
#             z_speed=z_speed,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_aspirate(
#         self,
#         volume: float,
#         location: Location = None,
#         pipette_cone: int = 1,
#         flow_rate: FlowRate = None,
#         blow_out_flow_rate: FlowRate = None,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.aspirate(
#             volume=volume,
#             location=location,
#             pipette_cone=pipette_cone,
#             flow_rate=flow_rate,
#             blow_out_flow_rate=blow_out_flow_rate,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_dispense(
#         self,
#         volume: float,
#         location: Location = None,
#         pipette_cone: int = 1,
#         flow_rate: FlowRate = None,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.dispense(
#             volume=volume,
#             location=location,
#             pipette_cone=pipette_cone,
#             flow_rate=flow_rate,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_mix(
#         self,
#         volume: float,
#         iteration: int,
#         location: Location = None,
#         pipette_cone: int = 1,
#         flow_rate: FlowRate = None,
#         blow_out_flow_rate: FlowRate = None,
#         delay: float = 0.0,
#         optimize: bool = False,
#     ):
#         await self._pipette_command.mix(
#             volume=volume,
#             iteration=iteration,
#             location=location,
#             pipette_cone=pipette_cone,
#             flow_rate=flow_rate,
#             blow_out_flow_rate=blow_out_flow_rate,
#             delay=delay,
#             optimize=optimize,
#         )

#     @log_func_args_async(log_before_func=logger.info, log_after_func=logger.info)
#     async def motion_blow_out(
#         self,
#         flow_rate: FlowRate = None,
#     ):
#         await self._pipette_command.blow_out(
#             flow_rate=flow_rate,
#         )

#     # axis api
#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_get_position(self, axis: Axis, floor_digit: int = 1):
#         return await self._axis_command.get_position(axis=axis, floor_digit=floor_digit)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_set_speed(self, axis: Axis, value: float):
#         return await self._axis_command.set_speed(axis=axis, value=value)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_set_accel(self, axis: Axis, value: float):
#         return await self._axis_command.set_accel(axis=axis, value=value)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_set_decel(self, axis: Axis, value: float):
#         return await self._axis_command.set_decel(axis=axis, value=value)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_enable(self, axis: Axis):
#         await self._axis_command.enable(axis=axis)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_disable(self, axis: Axis):
#         await self._axis_command.disable(axis=axis)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_stop(self, axis: Axis):
#         await self._axis_command.stop(axis=axis)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_home(self, axis: Axis):
#         await self._axis_command.home(axis=axis)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_jog(self, axis: Axis, value: float):
#         await self._axis_command.jog(axis=axis, value=value)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_step(self, axis: Axis, value: float):
#         await self._axis_command.step(axis=axis, value=value)

#     @log_func_args_async(log_after_func=logger.debug)
#     async def axis_move(self, axis: Axis, value: float):
#         await self._axis_command.move(axis=axis, value=value)

#     @log_func_args_async(log_before_func=logger.debug, log_after_func=logger.debug)
#     async def axis_wait_home_done(self, axis: Axis):
#         await self._axis_command.wait_home_done(axis=axis)

#     @log_func_args_async(log_before_func=logger.debug, log_after_func=logger.debug)
#     async def axis_wait_move_done(self, axis: Axis):
#         await self._axis_command.wait_move_done(axis=axis)


# async def build_api(
#     robot_context: RobotContext,
#     robot_controller: RobotController,
#     ip: str = "localhost",
#     port: int = 1234,
# ):
#     # log_func = logger.debug
#     log_func = logger.trace

#     tcp_server = TcpServer(name="tcp_server", log_func=log_func)
#     messenger_server = MessengerServer(
#         tcp_server=tcp_server,
#         robot_context=robot_context,
#         robot_controller=robot_controller,
#     )

#     async def on_connected():
#         # await robot.initialize()
#         pass

#     tcp_server.on_connected.append(on_connected)

#     async def on_received(msg: str):
#         pass

#     tcp_server.on_received.append(on_received)

#     async def send_api(msg: str):
#         # await tcp_server.send(msg)
#         pass

#     robot_controller._robot_command.send_api = send_api
#     await tcp_server.open(ip=ip, port=port)


# async def main():
#     logger.remove()
#     # logger.add(sys.stdout, level="TRACE")
#     logger.add(sys.stdout, level="INFO", backtrace=False)
#     # logger.add(sys.stdout, level="DEBUG", backtrace=True)
#     logger.add("logs/trace.log", level="TRACE")
#     logger.add("logs/debug.log", level="DEBUG")
#     logger.add("logs/info.log", level="INFO")
#     logger.add("logs/warning.log", level="WARNING")

#     ip = "localhost"
#     port = 1234

#     path = Path(required_dirs=["robot", "resources", "platform_autocal"])
#     robot_context = RobotContext(path=path)
#     robot_context.load_data(do_log=False)
#     robot_driver = RobotDriver(
#         robot_context=robot_context,
#     )
#     robot_command = RobotCommand(
#         robot_context=robot_context,
#         robot_driver=robot_driver,
#     )
#     dio_command = DIOCommand(
#         robot_context=robot_context,
#         robot_driver=robot_driver,
#     )
#     axis_command = AxisCommand(
#         robot_context=robot_context,
#         robot_driver=robot_driver,
#         robot_command=robot_command,
#     )
#     pipette_command = PipetteCommand(
#         robot_context=robot_context,
#         robot_command=robot_command,
#         axis_command=axis_command,
#     )
#     robot_controller = RobotController(
#         robot_context=robot_context,
#         robot_driver=robot_driver,
#         robot_command=robot_command,
#         dio_command=dio_command,
#         axis_command=axis_command,
#         pipette_command=pipette_command,
#     )
#     await build_api(
#         robot_context=robot_context,
#         robot_controller=robot_controller,
#         ip=ip,
#         port=port,
#     )
#     tasks = robot_controller.build_tasks()


# if __name__ == "__main__":
#     loop = asyncio.new_event_loop()
#     loop.set_debug(True)
#     loop.create_task(main())
#     loop.run_forever()



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import Color, RunStatus, TaskStatus, Plunger
from ablelabs.neon.utils.enum_types import Enum, auto, StrEnum, ItemGetable


class LedLamp(Enum):
    NONE = auto()
    RED = auto()
    GREEN = auto()
    BLUE = auto()


class Buzzer(Enum):
    DONE = auto()
    ERROR = auto()


class Axis(StrEnum):
    X = auto()
    Y1 = auto()
    Y2 = auto()
    Y3 = auto()
    Y4 = auto()
    Y5 = auto()
    Y6 = auto()
    Y7 = auto()
    Y8 = auto()
    Z1 = auto()
    Z2 = auto()
    Z3 = auto()
    Z4 = auto()
    Z5 = auto()
    Z6 = auto()
    Z7 = auto()
    Z8 = auto()
    P1 = auto()
    P2 = auto()
    P3 = auto()
    P4 = auto()
    P5 = auto()
    P6 = auto()
    P7 = auto()
    P8 = auto()

    Y = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.Y1[0]}{number}"))
    Z = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.Z1[0]}{number}"))
    P = ItemGetable[int](lambda number: Axis.from_str(f"{Axis.P1[0]}{number}"))


class LocationType(StrEnum):
    READY = auto()
    DECK = auto()
    TUBE_384 = auto()
    WASTE = auto()


class LocationReference(StrEnum):
    BOTTOM = auto()
    TOP = auto()
    TOP_JUST = auto()
    LIQUID = auto()  # for LLD only when aspirate


class Height(StrEnum):
    UP = auto()
    TOP = auto()
    TOP_JUST = auto()
    BOTTOM = auto()
    TIP_RACK_TOUCHED = auto()
    TIP_RACK_PRESSED = auto()
    WASTE_DROP_TIP = auto()
    WASTE_BLOW_OUT = auto()


class AssignmentStatus(StrEnum):
    NOT_ASSIGNED = auto()
    ASSIGNING = auto()
    ASSIGNED = auto()


class MotionStr(StrEnum):
    READY = auto()
    INITIALIZE = auto()
    PICK_UP_TIP = auto()
    ASPIRATE = auto()
    DISPENSE = auto()
    DROP_TIP = auto()


if __name__ == "__main__":
    y1 = Axis.Y[1]
    print(y1)
